from typing import List, Type

from packel.packet import Packet


class Protocol(object):
    def __init__(self, packet_classes: List[Type[Packet]]):
        self.packet_classes = packet_classes

    def serialize(self, packet: Packet) -> bytes:
        """Serializes a packet to bytes."""
        if not isinstance(packet, Packet):
            raise ValueError(f"'{type(packet)}' is not a packet type.")

        try:
            index: int = self.packet_classes.index(type(packet))
        except ValueError:
            raise ValueError(f"'{type(packet)}' packet type is not registered in the protocol.")

        return index.to_bytes(4, 'big') + packet.serialized()

    def deserialize(self, b: bytes) -> Packet:
        """Deserializes bytes to a packet.""" 
        index = int.from_bytes(b[:4], 'big')
        if index > len(self.packet_classes):
            raise ValueError(f"Packet type '{index}' exceeds number of registered protocol packet types.")

        return self.packet_classes[index].deserialize(b[4:])



from typing import List

from packel.types import PacketType


class Packet(object):
    def __init__(self, *args, **kwargs):
        """Initialize packel fields from the constructor."""
        self.serializable_fields: List[str] = []
        for field_name in dir(type(self)):
            if isinstance(getattr(type(self), field_name), PacketType):
                self.serializable_fields.append(field_name)
        self.serializable_fields.sort()

        for name, val in kwargs.items():
            if name in self.serializable_fields:
                setattr(self, name, val)

    def serialized(self) -> bytes:
        """Serializes a packet to bytes."""
        data: bytes = b''
        for name in self.serializable_fields:
            field = getattr(type(self), name)
            field_data = field.serialize(getattr(self, name))
            if field.is_dynamic_length:
                data = data + len(field_data).to_bytes(4, 'big')
            data = data + field_data

        return data

    @classmethod
    def deserialize(cls, b: bytes):
        """Deserializes bytes to a packet.""" 
        packet = cls()

        pos = 0
        for name in packet.serializable_fields:
            field = getattr(type(packet), name)
            if field.is_dynamic_length:
                bslice = b[pos:pos+4]
                pos = pos + 4
                field_len = int.from_bytes(bslice, 'big')
            else:
                field_len = len(field.serialize(getattr(packet, name)))

            bslice = b[pos:pos+field_len]
            pos = pos + field_len
            setattr(packet, name, field.deserialize(bslice))

        return packet



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="packel",
    version="1.0.0",
    author="kiwec",
    author_email="c.wolf@kiwec.net",
    description="Packet serialization/deserialization in a Pythonic way.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kiwec/packel",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from abc import ABC, abstractmethod
from typing import ClassVar, Type


class PacketType(ABC):
    is_dynamic_length: ClassVar[bool] = False
    type: ClassVar[Type]

    def __get__(self, instance, owner = None):
        # If the class is not initialized, return the descriptor to allow calling static methods.
        if instance is None:
            return self

        return self.value

    def __set__(self, instance, value) -> None:
        if not isinstance(value, self.type):
            raise AttributeError(f"Expected a {self.type}, but was given a {type(value)}.")
        self.value = value

    def __delete__(self, instance):
        self.value = self.default()

    def __init__(self):
        self.value = self.default()

    @staticmethod
    @abstractmethod
    def default():
        pass

    @staticmethod
    @abstractmethod
    def deserialize(b: bytes):
        pass

    @staticmethod
    @abstractmethod
    def serialize(v) -> bytes:
        pass


class Boolean(PacketType):
    is_dynamic_length: ClassVar[bool] = False
    type: ClassVar[Type] = bool

    @staticmethod
    def default() -> bool:
        return False

    @staticmethod
    def deserialize(b: bytes) -> bool:
        return int.from_bytes(b, 'big') == 1

    @staticmethod
    def serialize(v) -> bytes:
        x = 1 if v is True else 0
        return x.to_bytes(1, 'big')


class Bytes(PacketType):
    is_dynamic_length: ClassVar[bool] = True
    type: ClassVar[Type] = bytes

    @staticmethod
    def default() -> bytes:
        return b''

    @staticmethod
    def deserialize(b: bytes) -> bytes:
        return b

    @staticmethod
    def serialize(v) -> bytes:
        return v


class Int32(PacketType):
    is_dynamic_length: ClassVar[bool] = False
    type: ClassVar[Type] = int

    @staticmethod
    def default() -> int:
        return 0

    @staticmethod
    def deserialize(b: bytes) -> int:
        return int.from_bytes(b, 'big')

    @staticmethod
    def serialize(v) -> bytes:
        return v.to_bytes(4, 'big')


class Int64(PacketType):
    is_dynamic_length: ClassVar[bool] = False
    type: ClassVar[Type] = int

    @staticmethod
    def default() -> int:
        return 0

    @staticmethod
    def deserialize(b: bytes) -> int:
        return int.from_bytes(b, 'big')

    @staticmethod
    def serialize(v) -> bytes:
        return v.to_bytes(8, 'big')


class String(PacketType):
    is_dynamic_length: ClassVar[bool] = True
    type: ClassVar[Type] = str

    @staticmethod
    def default() -> str:
        return ''

    @staticmethod
    def deserialize(b: bytes) -> str:
        return b.decode('utf-8')

    @staticmethod
    def serialize(v) -> bytes:
        return v.encode('utf-8')
